from __future__ import print_function, absolute_import
import glob
import json
import os
import shutil
import platform
import sys
import subprocess
from distutils.cmd import Command
from distutils.errors import (
    CompileError,
    DistutilsExecError,
    DistutilsFileError,
    DistutilsPlatformError,
    DistutilsSetupError,
)
from subprocess import check_output

from .extension import RustExtension
from .utils import Binding, Strip, cpython_feature, get_rust_version


class build_rust(Command):
    """ Command for building rust crates via cargo. """

    description = "build Rust extensions (compile/link to build directory)"

    user_options = [
        (
            "inplace",
            "i",
            "ignore build-lib and put compiled extensions into the source "
            + "directory alongside your pure Python modules",
        ),
        ("debug", "d", "Force debug to true for all rust extensions "),
        ("release", "r", "Force debug to false for all rust extensions "),
        ("qbuild", None, "Force enable quiet option for all rust extensions "),
        (
            "build-temp",
            "t",
            "directory for temporary files (cargo 'target' directory) ",
        ),
    ]
    boolean_options = ["inplace", "debug", "release", "qbuild"]

    def initialize_options(self):
        self.extensions = ()
        self.inplace = None
        self.debug = None
        self.release = None
        self.qbuild = None
        self.build_temp = None
        self.plat_name = None

    def finalize_options(self):
        self.extensions = [
            ext
            for ext in self.distribution.rust_extensions
            if isinstance(ext, RustExtension)
        ]

        # Inherit settings from the `build_ext` command
        self.set_undefined_options(
            "build_ext",
            ("build_temp", "build_temp"),
            ("debug", "debug"),
            ("inplace", "inplace"),
        )

    def build_extension(self, ext):
        executable = ext.binding == Binding.Exec

        # Make sure that if pythonXX-sys is used, it builds against the current
        # executing python interpreter.
        bindir = os.path.dirname(sys.executable)

        env = os.environ.copy()
        env.update(
            {
                # disables rust's pkg-config seeking for specified packages,
                # which causes pythonXX-sys to fall back to detecting the
                # interpreter from the path.
                "PATH": os.path.join(bindir, os.environ.get("PATH", "")),
                "PYTHON_SYS_EXECUTABLE": sys.executable,
                "PYO3_PYTHON": sys.executable,
            }
        )

        # If we are on a 64-bit machine, but running a 32-bit Python, then
        # we'll target a 32-bit Rust build.
        # TODO: include --target for all platforms so env vars can't break the build
        target_triple = None
        target_args = []
        if platform.machine() == "AMD64" and self.plat_name == "win32":
            target_triple = "i686-pc-windows-msvc"
            target_args = ["--target", target_triple]

        # Find where to put the temporary build files created by `cargo`
        metadata_command = [
            "cargo",
            "metadata",
            "--manifest-path",
            ext.path,
            "--format-version",
            "1",
        ]
        # The decoding is needed for python 3.5 compatibility
        metadata = json.loads(check_output(metadata_command).decode("utf-8"))
        target_dir = metadata["target_directory"]

        if not os.path.exists(ext.path):
            raise DistutilsFileError(
                "Can not find rust extension project file: %s" % ext.path
            )

        features = set(ext.features)
        features.update(cpython_feature(binding=ext.binding))

        debug_build = ext.debug if ext.debug is not None else self.inplace
        debug_build = self.debug if self.debug is not None else debug_build
        if self.release:
            debug_build = False

        quiet = self.qbuild or ext.quiet

        # build cargo command
        feature_args = ["--features", " ".join(features)] if features else []

        if executable:
            args = (
                ["cargo", "build", "--manifest-path", ext.path]
                + feature_args
                + target_args
                + list(ext.args or [])
            )
            if not debug_build:
                args.append("--release")
            if quiet:
                args.append("-q")
            elif self.verbose:
                args.append("--verbose")

        else:
            args = (
                ["cargo", "rustc", "--lib", "--manifest-path", ext.path]
                + feature_args
                + target_args
                + list(ext.args or [])
            )
            if not debug_build:
                args.append("--release")
            if quiet:
                args.append("-q")
            elif self.verbose:
                args.append("--verbose")

            args.extend(["--", "--crate-type", "cdylib"])
            args.extend(ext.rustc_flags or [])

            # OSX requires special linker argument
            if sys.platform == "darwin":
                args.extend(
                    ["-C", "link-arg=-undefined", "-C", "link-arg=dynamic_lookup"]
                )

        if not quiet:
            print(" ".join(args), file=sys.stderr)

        if ext.native:
            env["RUSTFLAGS"] = "-C target-cpu=native"

        # Execute cargo
        try:
            output = subprocess.check_output(args, env=env)
        except subprocess.CalledProcessError as e:
            output = e.output
            if isinstance(output, bytes):
                output = e.output.decode("latin-1").strip()
            raise CompileError(
                "cargo failed with code: %d\n%s" % (e.returncode, output)
            )

        except OSError:
            raise DistutilsExecError(
                "Unable to execute 'cargo' - this package "
                "requires rust to be installed and cargo to be on the PATH"
            )

        if not quiet:
            if isinstance(output, bytes):
                output = output.decode("latin-1")
            if output:
                print(output, file=sys.stderr)

        # Find the shared library that cargo hopefully produced and copy
        # it into the build directory as if it were produced by build_ext.
        if debug_build:
            suffix = "debug"
        else:
            suffix = "release"

        # location of cargo compiled files
        artifactsdir = os.path.join(target_dir, target_triple or "", suffix)
        dylib_paths = []

        if executable:
            for name, dest in ext.target.items():
                if name:
                    path = os.path.join(artifactsdir, name)
                    if os.access(path, os.X_OK):
                        dylib_paths.append((dest, path))
                        continue
                    else:
                        raise DistutilsExecError(
                            "rust build failed; "
                            'unable to find executable "%s" in %s' % (name, target_dir)
                        )
                else:
                    # search executable
                    for name in os.listdir(artifactsdir):
                        path = os.path.join(artifactsdir, name)
                        if name.startswith(".") or not os.path.isfile(path):
                            continue

                        if os.access(path, os.X_OK):
                            dylib_paths.append((ext.name, path))
                            break

            if not dylib_paths:
                raise DistutilsExecError(
                    "rust build failed; unable to find executable in %s" % target_dir
                )
        else:
            if sys.platform == "win32" or sys.platform == "cygwin":
                dylib_ext = "dll"
            elif sys.platform == "darwin":
                dylib_ext = "dylib"
            else:
                dylib_ext = "so"

            wildcard_so = "*{}.{}".format(ext.get_lib_name(), dylib_ext)

            try:
                dylib_paths.append(
                    (
                        ext.name,
                        next(glob.iglob(os.path.join(artifactsdir, wildcard_so))),
                    )
                )
            except StopIteration:
                raise DistutilsExecError(
                    "rust build failed; unable to find any %s in %s"
                    % (wildcard_so, artifactsdir)
                )

        # Ask build_ext where the shared library would go if it had built it,
        # then copy it there.
        build_ext = self.get_finalized_command("build_ext")
        build_ext.inplace = self.inplace

        for target_fname, dylib_path in dylib_paths:
            if not target_fname:
                target_fname = os.path.basename(
                    os.path.splitext(os.path.basename(dylib_path)[3:])[0]
                )

            if executable:
                ext_path = build_ext.get_ext_fullpath(target_fname)
                # remove .so extension
                ext_path, _ = os.path.splitext(ext_path)
                # remove python3 extension (i.e. cpython-36m)
                ext_path, _ = os.path.splitext(ext_path)

                ext.install_script(ext_path)
            else:
                ext_path = build_ext.get_ext_fullpath(target_fname)

            try:
                os.makedirs(os.path.dirname(ext_path))
            except OSError:
                pass

            shutil.copyfile(dylib_path, ext_path)

            if sys.platform != "win32" and not debug_build:
                args = []
                if ext.strip == Strip.All:
                    args.append("-x")
                elif ext.strip == Strip.Debug:
                    args.append("-S")

                if args:
                    args.insert(0, "strip")
                    args.append(ext_path)
                    try:
                        output = subprocess.check_output(args, env=env)
                    except subprocess.CalledProcessError as e:
                        pass

            # executables and win32(cygwin)-dll's need X bits
            if executable or sys.platform == "win32" or sys.platform == "cygwin":
                mode = os.stat(ext_path).st_mode
                mode |= (mode & 0o444) >> 2  # copy R bits to X
                os.chmod(ext_path, mode)

    def run(self):
        if not self.extensions:
            return

        all_optional = all(ext.optional for ext in self.extensions)
        try:
            version = get_rust_version()
        except DistutilsPlatformError as e:
            if not all_optional:
                raise
            else:
                print(str(e))
                return

        for ext in self.extensions:
            try:
                rust_version = ext.get_rust_version()
                if rust_version is not None and version not in rust_version:
                    raise DistutilsPlatformError(
                        "Rust %s does not match extension requirement %s"
                        % (version, ext.rust_version)
                    )

                self.build_extension(ext)
            except (
                DistutilsSetupError,
                DistutilsFileError,
                DistutilsExecError,
                DistutilsPlatformError,
                CompileError,
            ) as e:
                if not ext.optional:
                    raise
                else:
                    print("Build optional Rust extension %s failed." % ext.name)
                    print(str(e))



# coding: utf-8
import glob
import os
import string

try:
    import configparser
except ImportError:
    import ConfigParser as configparser

import setuptools

from distutils import log
from distutils.command.build import build

from .extension import RustExtension


__all__ = ["tomlgen"]


class tomlgen_rust(setuptools.Command):

    description = "Generate `Cargo.toml` for rust extensions"

    user_options = [
        (str("force"), str("f"), str("overwrite existing files if any")),
        (
            str("create-workspace"),
            str("w"),
            str("create a workspace file at the root of the project"),
        ),
        (
            str("no-config"),
            str("C"),
            str("do not create a `.cargo/config` file when generating a workspace"),
        ),
    ]

    boolean_options = [str("create_workspace"), str("force")]

    def initialize_options(self):

        self.dependencies = None
        self.authors = None
        self.create_workspace = None
        self.no_config = None
        self.force = None

        # use the build command to find build directories
        self.build = build(self.distribution)

        # parse config files
        self.cfg = configparser.ConfigParser()
        self.cfg.read(self.distribution.find_config_files())

    def finalize_options(self):

        # Finalize previous commands
        self.distribution.finalize_options()
        self.build.ensure_finalized()

        # Shortcuts
        self.extensions = self.distribution.rust_extensions
        self.workspace = os.path.abspath(
            os.path.dirname(self.distribution.script_name) or "."
        )

        # Build list of authors
        if self.authors is not None:
            sep = "\n" if "\n" in self.authors.strip() else ","
            self.authors = "[{}]".format(
                ", ".join(author.strip() for author in self.authors.split(sep))
            )
        else:
            self.authors = '["{} <{}>"]'.format(
                self.distribution.get_author(),
                self.distribution.get_author_email().strip("\"'"),
            )

    def run(self):

        # Create a `Cargo.toml` for each extension
        for ext in self.extensions:
            toml = self.build_cargo_toml(ext)
            if not os.path.exists(ext.path) or self.force:
                log.info("creating 'Cargo.toml' for '%s'", ext.name)
                with open(ext.path, "w") as manifest:
                    toml.write(manifest)
            else:
                log.warn("skipping 'Cargo.toml' for '%s' -- already exists", ext.name)

        # Create a `Cargo.toml` for the project workspace
        if self.create_workspace and self.extensions:
            toml = self.build_workspace_toml()
            toml_path = os.path.join(self.workspace, "Cargo.toml")
            if not os.path.exists(toml_path) or self.force:
                log.info("creating 'Cargo.toml' for workspace")
                with open(toml_path, "w") as manifest:
                    toml.write(manifest)
            else:
                log.warn("skipping 'Cargo.toml' for workspace -- already exists")

        # Create a `.cargo/config` file
        if self.create_workspace and self.extensions and not self.no_config:

            dist = self.distribution
            targetdir = os.path.join(self.build.build_temp, dist.get_name())
            cfgdir = os.path.abspath(
                os.path.join(os.getcwd(), dist.script_name, "..", ".cargo")
            )

            if not os.path.exists(os.path.join(cfgdir, "config")) or self.force:
                if not os.path.exists(cfgdir):
                    os.makedirs(cfgdir)
                with open(os.path.join(cfgdir, "config"), "w") as config:
                    log.info("creating '.cargo/config' for workspace")

                    config.write("[build]\n")
                    config.write(
                        'target-dir = "{}"\n'.format(os.path.relpath(targetdir))
                    )

            else:
                log.warn("skipping '.cargo/config' -- already exists")

    def build_cargo_toml(self, ext):

        # Shortcuts
        quote = '"{}"'.format
        dist = self.distribution

        # Use a ConfigParser object to build a TOML file (hackish)
        toml = configparser.ConfigParser()

        # The directory where the extension's manifest is located
        tomldir = os.path.dirname(ext.path)

        # If the RustExtension was not created by `find_rust_extensions`
        # the `lib.rs` file is expected to be located near `Cargo.toml`
        if not hasattr(ext, "libfile"):
            ext.libfile = ext.path.replace("Cargo.toml", "lib.rs")

        # Create a small package section
        toml.add_section("package")
        toml.set("package", "name", quote(ext.name.replace('.', '-')))
        toml.set("package", "version", quote(dist.get_version()))
        toml.set("package", "authors", self.authors)
        toml.set("package", "publish", "false")

        # Add the relative path to the workspace if any
        if self.create_workspace:
            path_to_workspace = os.path.relpath(self.workspace, tomldir)
            toml.set("package", "workspace", quote(path_to_workspace))

        # Create a small lib section
        toml.add_section("lib")
        toml.set("lib", "crate-type", '["cdylib"]')
        toml.set("lib", "name", quote(_slugify(ext.name)))
        toml.set("lib", "path", quote(os.path.relpath(ext.libfile, tomldir)))

        # Find dependencies within the `setup.cfg` file of the project
        toml.add_section("dependencies")
        for dep, options in self.iter_dependencies(ext):
            toml.set("dependencies", dep, options)

        return toml

    def build_workspace_toml(self):

        # Find all members of the workspace
        members = [
            os.path.dirname(os.path.relpath(ext.path)) for ext in self.extensions
        ]
        members = ['"{}"'.format(m) for m in members]

        # Create the `Cargo.toml` content using a ConfigParser
        toml = configparser.ConfigParser()
        toml.add_section("workspace")
        toml.set("workspace", "members", "[{}]".format(", ".join(members)))

        return toml

    def iter_dependencies(self, ext=None):

        command = self.get_command_name()

        # global dependencies
        sections = ["{}.dependencies".format(command)]

        # extension-specific dependencies
        if ext is not None:
            sections.append("{}.dependencies.{}".format(command, ext.name))

        for section in sections:
            if self.cfg.has_section(section):
                for dep, options in self.cfg.items(section):
                    yield dep, options


def _slugify(name):
    allowed = set(string.ascii_letters + string.digits + "_")
    slug = [char if char in allowed else "_" for char in name]
    return "".join(slug)


def find_rust_extensions(*directories, **kwargs):
    """Attempt to find Rust extensions in given directories.

    This function will recurse through the directories in the given
    directories, to find a name whose name is ``libfile``. When such
    a file is found, an extension is created, expecting the cargo
    manifest file (``Cargo.toml``) to be next to that file. The
    extension destination will be deduced from the name of the
    directory where that ``libfile`` is contained.

    Arguments:
        directories (list, *optional*): a list of directories to walk
            through recursively to find extensions. If none are given,
            then the current directory will be used instead.

    Keyword Arguments:
        libfile (str): the name of the file to look for when searching
            for Rust extensions. Defaults to ``lib.rs``, but might be
            changed to allow defining more *Pythonic* filenames
            (like ``__init__.rs``)!

    Note:
        All other keyword arguments will be directly passed to the
        `RustExtension` instance created when an extension is found.
        One may be interested in passing ``bindings`` and ``strip``
        options::

            >>> import setuptools_rust as rust
            >>> rust.find_rust_extensions(binding=rust.Binding.PyO3)

    Example:

        Consider the following project::

            lib/
             └ mylib/
                 └ rustext/
                     ├ lib.rs
                     ├  ...
                     └  Cargo.toml
            setup.py

        There is only one extension that can be found in the ``lib``
        module::

            >>> import setuptools_rust as rust
            >>> for ext in rust.find_rust_extensions("lib"):
            ...     print(ext.name, "=>", ext.path)
            lib.mylib.rustext => lib/mylib/rustext/Cargo.toml
    """

    # Get the file used to mark a Rust extension
    libfile = kwargs.get("libfile", "lib.rs")

    # Get the directories to explore
    directories = directories or [os.getcwd()]

    extensions = []
    for directory in directories:
        for base, dirs, files in os.walk(directory):
            if libfile in files:
                dotpath = os.path.relpath(base).replace(os.path.sep, ".")
                tomlpath = os.path.join(base, "Cargo.toml")
                ext = RustExtension(dotpath, tomlpath, **kwargs)
                ext.libfile = os.path.join(base, libfile)
                extensions.append(ext)

    return extensions



from __future__ import print_function, absolute_import

from .build import build_rust
from .check import check_rust
from .clean import clean_rust
from .extension import RustExtension
from .test import test_rust
from .tomlgen import tomlgen_rust, find_rust_extensions
from .utils import Binding, Strip

__all__ = (
    "RustExtension",
    "Binding",
    "Strip",
    "check_rust",
    "clean_rust",
    "build_rust",
    "test_rust",
)



from distutils import log
from distutils.command.check import check
from distutils.command.clean import clean
from distutils.command.install import install

from setuptools.command.build_ext import build_ext

try:
    from wheel.bdist_wheel import bdist_wheel
except ImportError:
    bdist_wheel = None


def add_rust_extension(dist):
    build_ext_base_class = dist.cmdclass.get('build_ext', build_ext)

    class build_ext_rust_extension(build_ext_base_class):
        def run(self):
            if self.distribution.rust_extensions:
                log.info("running build_rust")
                build_rust = self.get_finalized_command("build_rust")
                build_rust.inplace = self.inplace
                build_rust.plat_name = self.plat_name
                build_rust.run()

            build_ext_base_class.run(self)
    dist.cmdclass['build_ext'] = build_ext_rust_extension

    clean_base_class = dist.cmdclass.get('clean', clean)

    class clean_rust_extension(clean_base_class):
        def run(self):
            clean_base_class.run(self)
            if not self.dry_run:
                self.run_command("clean_rust")
    dist.cmdclass['clean'] = clean_rust_extension

    check_base_class = dist.cmdclass.get('check', check)

    class check_rust_extension(check_base_class):
        def run(self):
            check_base_class.run(self)
            self.run_command("check_rust")
    dist.cmdclass["check"] = check_rust_extension

    install_base_class = dist.cmdclass.get('install', install)

    # this is required because, install directly access distribution's
    # ext_modules attr to check if dist has ext modules
    class install_rust_extension(install_base_class):
        def finalize_options(self):
            ext_modules = self.distribution.ext_modules

            # all ext modules
            mods = []
            if self.distribution.ext_modules:
                mods.extend(self.distribution.ext_modules)
            if self.distribution.rust_extensions:
                mods.extend(self.distribution.rust_extensions)

                scripts = []
                for ext in self.distribution.rust_extensions:
                    scripts.extend(ext.entry_points())

                if scripts:
                    if not self.distribution.entry_points:
                        self.distribution.entry_points = {"console_scripts": scripts}
                    else:
                        ep_scripts = self.distribution.entry_points.get("console_scripts")
                        if ep_scripts:
                            for script in scripts:
                                if script not in ep_scripts:
                                    ep_scripts.append(scripts)
                        else:
                            ep_scripts = scripts

                        self.distribution.entry_points["console_scripts"] = ep_scripts

            self.distribution.ext_modules = mods

            install_base_class.finalize_options(self)

            # restore ext_modules
            self.distribution.ext_modules = ext_modules
    dist.cmdclass["install"] = install_rust_extension

    if bdist_wheel is not None:
        bdist_wheel_base_class = dist.cmdclass.get("bdist_wheel", bdist_wheel)

        # this is for console entries
        class bdist_wheel_rust_extension(bdist_wheel_base_class):
            def finalize_options(self):
                scripts = []
                for ext in self.distribution.rust_extensions:
                    scripts.extend(ext.entry_points())

                if scripts:
                    if not self.distribution.entry_points:
                        self.distribution.entry_points = {"console_scripts": scripts}
                    else:
                        ep_scripts = self.distribution.entry_points.get("console_scripts")
                        if ep_scripts:
                            for script in scripts:
                                if script not in ep_scripts:
                                    ep_scripts.append(scripts)
                        else:
                            ep_scripts = scripts

                        self.distribution.entry_points["console_scripts"] = ep_scripts

                bdist_wheel_base_class.finalize_options(self)
        dist.cmdclass["bdist_wheel"] = bdist_wheel_rust_extension



def rust_extensions(dist, attr, value):
    assert attr == "rust_extensions"

    orig_has_ext_modules = dist.has_ext_modules
    dist.has_ext_modules = lambda: (
        orig_has_ext_modules() or bool(dist.rust_extensions)
    )

    if dist.rust_extensions:
        add_rust_extension(dist)



from __future__ import print_function, absolute_import
import os
import sys
import subprocess
from distutils.cmd import Command
from distutils.errors import CompileError, DistutilsFileError, DistutilsExecError

import semantic_version

from .extension import RustExtension
from .utils import cpython_feature, get_rust_version

MIN_VERSION = semantic_version.Spec(">=1.15")


class test_rust(Command):
    """ Run cargo test"""

    description = "test rust extensions"

    user_options = []

    def initialize_options(self):
        self.extensions = ()

    def finalize_options(self):
        self.extensions = [
            ext
            for ext in self.distribution.rust_extensions
            if isinstance(ext, RustExtension)
        ]

    def run(self):
        if not self.extensions:
            return

        version = get_rust_version()
        if version not in MIN_VERSION:
            print(
                "Rust version mismatch: required rust%s got rust%s"
                % (MIN_VERSION, version)
            )
            return

        # Make sure that if pythonXX-sys is used, it builds against the current
        # executing python interpreter.
        bindir = os.path.dirname(sys.executable)

        env = os.environ.copy()
        env.update(
            {
                # disables rust's pkg-config seeking for specified packages,
                # which causes pythonXX-sys to fall back to detecting the
                # interpreter from the path.
                "PYTHON_2.7_NO_PKG_CONFIG": "1",
                "PATH": bindir + os.pathsep + os.environ.get("PATH", ""),
            }
        )

        for ext in self.extensions:
            if not os.path.exists(ext.path):
                raise DistutilsFileError(
                    "Can not file rust extension project file: %s" % ext.path
                )

            features = set(ext.features)
            features.update(cpython_feature(ext=False, binding=ext.binding))

            # test cargo command
            feature_args = ["--features", " ".join(features)] if features else []
            args = (
                ["cargo", "test", "--manifest-path", ext.path]
                + feature_args
                + list(ext.args or [])
            )

            # Execute cargo command
            print(" ".join(args))
            try:
                subprocess.check_output(args, env=env)
            except subprocess.CalledProcessError as e:
                raise CompileError(
                    "cargo failed with code: %d\n%s"
                    % (e.returncode, e.output.decode("utf-8"))
                )
            except OSError:
                raise DistutilsExecError(
                    "Unable to execute 'cargo' - this package "
                    "requires rust to be installed and "
                    "cargo to be on the PATH"
                )
            else:
                print("Test has been completed for '%s' extension" % ext.name)



from __future__ import print_function, absolute_import
import sys
import subprocess
from distutils.errors import DistutilsPlatformError

import semantic_version


class Binding:
    """
    Binding Options
    """
    # https://github.com/PyO3/PyO3
    PyO3 = 0
    # https://github.com/dgrunwald/rust-cpython
    RustCPython = 1
    # Bring your own binding
    NoBinding = 2
    # Build executable
    Exec = 3


class Strip:
    """
    Strip Options
    """
    # do not strip symbols
    No = 0
    # strip debug symbols
    Debug = 1
    # strip all symbos
    All = 2


def cpython_feature(ext=True, binding=Binding.PyO3):
    version = sys.version_info

    if binding in (Binding.NoBinding, Binding.Exec):
        return ()
    elif binding is Binding.PyO3:
        if version > (3, 5):
            if ext:
                return {"pyo3/extension-module"}
            else:
                return {}
        else:
            raise DistutilsPlatformError("Unsupported python version: %s" % sys.version)
    elif binding is Binding.RustCPython:
        if (3, 3) < version:
            if ext:
                return {"cpython/python3-sys", "cpython/extension-module"}
            else:
                return {"cpython/python3-sys"}
        else:
            raise DistutilsPlatformError("Unsupported python version: %s" % sys.version)
    else:
        raise DistutilsPlatformError('Unknown Binding: "{}" '.format(binding))


def get_rust_version():
    try:
        output = subprocess.check_output(["rustc", "-V"])
        if isinstance(output, bytes):
            output = output.decode("latin-1")
        return semantic_version.Version(output.split(" ")[1], partial=True)
    except (subprocess.CalledProcessError, OSError):
        raise DistutilsPlatformError("Can not find Rust compiler")
    except Exception as exc:
        raise DistutilsPlatformError("Can not get rustc version: %s" % str(exc))



from __future__ import print_function, absolute_import
import os
import re
from distutils.errors import DistutilsSetupError
from .utils import Binding, Strip

import semantic_version


class RustExtension:
    """Just a collection of attributes that describes an rust extension
    module and everything needed to build it

    Instance attributes:
      target : string
        the full name of the extension, including any packages -- ie.
        *not* a filename or pathname, but Python dotted name
      path : string
        path to the cargo.toml manifest file
      args : [string]
        a list of extra argumenents to be passed to cargo.
      features : [string]
        a list of features to also build
      rust_version : string
        rust compiler version
      quiet : bool
        If True, doesn't echo cargo's output.
      debug : bool
        Controls whether --debug or --release is passed to cargo. If set to
        None then build type is auto-detect. Inplace build is debug build
        otherwise release. Default: None
      binding : setuptools_rust.Binding
        Controls which python binding is in use.
        Binding.PyO3 uses PyO3
        Binding.RustCPython uses Rust CPython.
        Binding.NoBinding uses no binding.
        Binding.Exec build executable.
      strip : setuptools_rust.Strip
        Strip symbols from final file. Does nothing for debug build.
        * Strip.No - do not strip symbols
        * Strip.Debug - strip debug symbols
        * Strip.All - strip all symbols
      script : bool
        Generate console script for executable if `Binding.Exec` is used.
      native : bool
        Build extension or executable with "target-cpu=native"
      optional : bool
        if it is true, a build failure in the extension will not abort the
        build process, but instead simply not install the failing extension.
    """

    def __init__(
        self,
        target,
        path="Cargo.toml",
        args=None,
        features=None,
        rustc_flags=None,
        rust_version=None,
        quiet=False,
        debug=None,
        binding=Binding.PyO3,
        strip=Strip.No,
        script=False,
        native=False,
        optional=False,
    ):
        if isinstance(target, dict):
            name = "; ".join("%s=%s" % (key, val) for key, val in target.items())
        else:
            name = target
            target = {"": target}

        self.name = name
        self.target = target
        self.args = args
        self.rustc_flags = rustc_flags
        self.binding = binding
        self.rust_version = rust_version
        self.quiet = quiet
        self.debug = debug
        self.strip = strip
        self.script = script
        self.native = native
        self.optional = optional

        if features is None:
            features = []

        self.features = [s.strip() for s in features]

        # get relative path to Cargo manifest file
        path = os.path.relpath(path)
        # file = sys._getframe(1).f_globals.get('__file__')
        # if file:
        #     dirname = os.path.dirname(file)
        #     print(dirname)
        #     if dirname:
        #         cwd = os.getcwd()
        #         os.chdir(dirname)
        #         path = os.path.abspath(path)
        #         os.chdir(cwd)

        self.path = path

    def get_lib_name(self):
        """ Parse Cargo.toml to get the name of the shared library. """
        # We import in here to make sure the the setup_requires are already installed
        import toml

        cfg = toml.load(self.path)
        name = cfg.get("lib", {}).get("name")
        if name is None:
            name = cfg.get("package", {}).get("name")
        if name is None:
            raise Exception(
                "Can not parse library name from Cargo.toml. "
                "Cargo.toml missing value for 'name' key "
                "in both the [package] section and the [lib] section"
            )
        name = re.sub(r"[./\\-]", "_", name)
        return name

    def get_rust_version(self):
        if self.rust_version is None:
            return None
        try:
            return semantic_version.Spec(self.rust_version)
        except ValueError:
            raise DistutilsSetupError(
                "Can not parse rust compiler version: %s", self.rust_version
            )

    def entry_points(self):
        entry_points = []
        if self.script and self.binding == Binding.Exec:
            for name, mod in self.target.items():
                base_mod, name = mod.rsplit(".")
                script = "%s=%s.%s:run" % (name, base_mod, "_gen_%s" % name)
                entry_points.append(script)

        return entry_points

    def install_script(self, ext_path):
        if self.script and self.binding == Binding.Exec:
            dirname, name = os.path.split(ext_path)
            file = os.path.join(dirname, "_gen_%s.py" % name)
            with open(file, "w") as f:
                f.write(TMPL.format({"name": name}))


TMPL = """from __future__ import absolute_import, print_function

import os
import sys


def run():
    path = os.path.split(__file__)[0]
    name = os.path.split(sys.argv[0])[1]
    file = os.path.join(path, name)
    if os.path.isfile(file):
        os.execv(file, sys.argv)
    else:
        print("Can not execute '%s'" % name)
"""



from __future__ import print_function, absolute_import
import sys
import subprocess
from distutils.cmd import Command

from .extension import RustExtension


class clean_rust(Command):
    """ Clean rust extensions. """

    description = "clean rust extensions (compile/link to build directory)"

    def initialize_options(self):
        self.extensions = ()
        self.inplace = False

    def finalize_options(self):
        self.extensions = [
            ext
            for ext in self.distribution.rust_extensions
            if isinstance(ext, RustExtension)
        ]

    def run(self):
        if not self.extensions:
            return

        for ext in self.extensions:
            # build cargo command
            args = ["cargo", "clean", "--manifest-path", ext.path]

            if not ext.quiet:
                print(" ".join(args), file=sys.stderr)

            # Execute cargo command
            try:
                subprocess.check_output(args)
            except:
                pass



from __future__ import print_function, absolute_import
import os
import sys
import subprocess
from distutils.cmd import Command
from distutils.errors import (
    CompileError,
    DistutilsFileError,
    DistutilsExecError,
    DistutilsPlatformError,
)

import semantic_version

from .extension import RustExtension
from .utils import cpython_feature, get_rust_version

MIN_VERSION = semantic_version.Spec(">=1.16")


class check_rust(Command):
    """ Run rust check"""

    description = "check rust extensions"

    def initialize_options(self):
        self.extensions = ()

    def finalize_options(self):
        self.extensions = [
            ext
            for ext in self.distribution.rust_extensions
            if isinstance(ext, RustExtension)
        ]

    def run(self):
        if "sdist" in self.distribution.commands:
            return

        if not self.extensions:
            return

        all_optional = all(ext.optional for ext in self.extensions)
        try:
            version = get_rust_version()
        except DistutilsPlatformError as e:
            if not all_optional:
                raise
            else:
                print(str(e))
                return
        if version not in MIN_VERSION:
            print(
                "Rust version mismatch: required rust%s got rust%s"
                % (MIN_VERSION, version)
            )
            return

        # Make sure that if pythonXX-sys is used, it builds against the current
        # executing python interpreter.
        bindir = os.path.dirname(sys.executable)

        env = os.environ.copy()
        env.update(
            {
                # disables rust's pkg-config seeking for specified packages,
                # which causes pythonXX-sys to fall back to detecting the
                # interpreter from the path.
                "PYTHON_2.7_NO_PKG_CONFIG": "1",
                "PATH": bindir + os.pathsep + os.environ.get("PATH", ""),
            }
        )

        for ext in self.extensions:
            try:
                if not os.path.exists(ext.path):
                    raise DistutilsFileError(
                        "Can not file rust extension project file: %s" % ext.path
                    )

                features = set(ext.features)
                features.update(cpython_feature(binding=ext.binding))

                # check cargo command
                feature_args = ["--features", " ".join(features)] if features else []
                args = (
                    ["cargo", "check", "--manifest-path", ext.path]
                    + feature_args
                    + list(ext.args or [])
                )

                # Execute cargo command
                try:
                    subprocess.check_output(args)
                except subprocess.CalledProcessError as e:
                    raise CompileError(
                        "cargo failed with code: %d\n%s"
                        % (e.returncode, e.output.decode("utf-8"))
                    )
                except OSError:
                    raise DistutilsExecError(
                        "Unable to execute 'cargo' - this package "
                        "requires rust to be installed and "
                        "cargo to be on the PATH"
                    )
                else:
                    print("Extension '%s' checked" % ext.name)
            except (DistutilsFileError, DistutilsExecError, CompileError) as e:
                if not ext.optional:
                    raise
                else:
                    print("Check optional Rust extension %s failed." % ext.name)
                    print(str(e))
